# -*- coding: UTF-8 -*-
import glob
import os
import sys
import json

import cv2
import numpy as np

import rospy
import rosbag
# from cv_bridge import CvBridge
from sensor_msgs.msg import CameraInfo, Image

from collections import deque
from scipy.spatial.transform import Rotation as R
import matplotlib as mpl
import matplotlib.cm as cm


class CvBridge:
    def __init__(self):
        pass
    
    @staticmethod
    def imgmsg_to_cv2(img_msg):
        if img_msg.encoding == "bgr8":
            dtype = np.dtype("uint8") # Hardcode to 8 bits...
            dtype = dtype.newbyteorder('>' if img_msg.is_bigendian else '<')
            image_opencv = np.ndarray(shape=(img_msg.height, img_msg.width, 3), # and three channels of data. Since OpenCV works with bgr natively, we don't need to reorder the channels.
                            dtype=dtype, buffer=img_msg.data)
        elif img_msg.encoding == "mono8":
            dtype = np.dtype("uint8") # Hardcode to 8 bits...
            dtype = dtype.newbyteorder('>' if img_msg.is_bigendian else '<')
            # print(len(img_msg.data))
            # print((img_msg.height, img_msg.width, 1))
            image_opencv = np.ndarray(shape=(img_msg.height, img_msg.width, 1), # and three channels of data. Since OpenCV works with bgr natively, we don't need to reorder the channels.
                            dtype=dtype, buffer=img_msg.data)
        elif img_msg.encoding == "mono16":
            dtype = np.dtype("uint16") # Hardcode to 8 bits...
            dtype = dtype.newbyteorder('>' if img_msg.is_bigendian else '<')
            image_opencv = np.ndarray(shape=(img_msg.height, img_msg.width, 1), # and three channels of data. Since OpenCV works with bgr natively, we don't need to reorder the channels.
                            dtype=dtype, buffer=img_msg.data)
        # If the byt order is different between the message and the system.
        if img_msg.is_bigendian == (sys.byteorder == 'little'):
            image_opencv = image_opencv.byteswap().newbyteorder()
        return image_opencv

    @staticmethod
    def cv2_to_imgmsg(cv_image, encoding):
        img_msg = Image()
        img_msg.height = cv_image.shape[0]
        img_msg.width = cv_image.shape[1]
        img_msg.encoding = encoding
        img_msg.is_bigendian = 0
        img_msg.data = cv_image.tostring()
        img_msg.step = len(img_msg.data) // img_msg.height # That double line is actually integer division, not a comment
        return img_msg


def dummy_disparity(left, right):
    """Return dummpy disparity
    """
    return np.zeros_like(left, dtype=np.float32)

def search_nearest(buffer, target_t):
    """Search for nearest timestamps
    """
    min_space = 1e9
    min_res = None
    for msg, t in buffer:
        space = abs(target_t.to_sec() - t.to_sec())
        if space < min_space:
            min_space = space
            min_res = (msg, t)
    return min_res

def search_nearest_name(img_list, target_t):
    """Search for nearest timestamps
    """
    timestr = "%.6f" % target_t.to_sec()
    min_space = 1e9
    min_res = None
    for path in img_list:
        t = float(path.split('/')[-1].split('.')[0].replace('_', '.'))
        space = abs(float(timestr) - t)
        if space < min_space:
            min_space = space
            min_res = path
    return min_res

def dir_inference(stereo_dir, timestamp_ns):
    disp_path = glob.glob(stereo_dir + f'/{timestamp_ns}*.tiff')[0]
    disp_img = cv2.imread(disp_path, cv2.IMREAD_UNCHANGED)
    return disp_img

def parser_stereo_param(file_path):
    """parser stereo_param_0.yml and init maps for rectify.
    """
    assert os.path.exists(file_path)
    fs = cv2.FileStorage(file_path, cv2.FILE_STORAGE_READ)

    img_height = int(fs.getNode("image_height").real())
    img_width = int(fs.getNode("image_width").real())

    camera_left = fs.getNode("camera_matrix_1").mat()
    camera_right = fs.getNode("camera_matrix_2").mat()

    dist_left = fs.getNode("distortion_1").mat()
    dist_right = fs.getNode("distortion_2").mat()

    t_stereo = fs.getNode("translation_vector").mat()

    euler = []
    for i in range(3):
        euler.append(fs.getNode("euler").at(i).real())
    r = R.from_euler('xyz', euler, degrees=True)
    rm = r.as_matrix()

    # print(img_height, img_width, camera_left, camera_right, dist_left, dist_right, t_stereo, rm)

    # init map
    size = (img_width, img_height)
    R1, R2, P1, P2, Q, validPixROI1, validPixROI2 = cv2.stereoRectify(camera_left, dist_left,
                                                                      camera_right, dist_right, size, rm, t_stereo,
                                                                      flags=cv2.CALIB_ZERO_DISPARITY, alpha=0)
    left_mapx, left_mapy = cv2.initUndistortRectifyMap(camera_left, dist_left, R1, P1, size, cv2.CV_32FC1)
    right_mapx, right_mapy = cv2.initUndistortRectifyMap(camera_right, dist_right, R2, P2, size, cv2.CV_32FC1)

    return (left_mapx, left_mapy, right_mapx, right_mapy, P2)


def remap(left, right, maps):
    """Remap left and right images
    """
    left_rectified = cv2.remap(left, maps[0], maps[1], cv2.INTER_LINEAR)
    right_rectified = cv2.remap(right, maps[2], maps[3], cv2.INTER_LINEAR)
    return left_rectified, right_rectified


class BagData:
    def __init__(self, bag_file, start_time=None, param_file='/home/hadoop-uavcvml/cephfs/data/yaoxinghua/data/bags/safe_landing2/stereo_param_0-20210315.yml', out_dir='bag_result'):
        self.bag_in = rosbag.Bag(bag_file, "r")
        self.maps = parser_stereo_param(param_file)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        
        if start_time:
            self.start_time = rospy.Time(self.bag_in.get_start_time() + float(start_time))
        else:
            self.start_time = rospy.Time(self.bag_in.get_start_time())
        
        self.rectified = False
        self.left_queue = deque(maxlen=3)
        self.right_queue = deque(maxlen=3)
        self.bridge = CvBridge()
        save_bag = os.path.join(out_dir, os.path.basename(bag_file)[:-4] + '_dnn.bag')
        self.bag_out = rosbag.Bag(save_bag, "w")
        
    def get_stereo(self):
        for topic, msg, t in self.bag_in.read_messages(start_time=self.start_time):
            timestr = msg.header.stamp
            if topic == '/mo_camera/left_image':
                left_image = self.bridge.imgmsg_to_cv2(msg)
                if left_image.shape[2] == 3:
                    left_image = cv2.cvtColor(left_image,cv2.COLOR_BGR2GRAY)
                self.left_queue.append((left_image, timestr))
            elif topic == '/mo_camera/right_image':
                right_image = self.bridge.imgmsg_to_cv2(msg)
                if right_image.shape[2] == 3:
                    right_image = cv2.cvtColor(right_image,cv2.COLOR_BGR2GRAY)
                self.right_queue.append((right_image,timestr))
            elif topic == '/mo_camera/left_rectified':
                left_image = self.bridge.imgmsg_to_cv2(msg)
                left_image = cv2.resize(left_image, (1280, 800))
                self.left_queue.append((left_image,timestr))
                self.rectified = True
            elif topic == '/mo_camera/right_rectified':
                right_image = self.bridge.imgmsg_to_cv2(msg)
                right_image = cv2.resize(right_image, (1280, 800))
                self.right_queue.append((right_image,timestr))
            else:
                # skip disp
                if 'dsp_image' in topic: continue
                # directly save to bag
                self.bag_out.write(topic, msg, t=t)
                continue
            
            if len(self.left_queue) == 0 or len(self.right_queue) == 0:
                continue
            right, right_time = self.right_queue.pop()
            left, left_time = search_nearest(self.left_queue, right_time)
            print(left_time,right_time)
            if left_time != right_time: continue
            if not self.rectified:
                self.left_rec, self.right_rec = remap(left, right, self.maps)
            else:
                self.left_rec, self.right_rec = left, right
            self.msg = msg
            self.t = t
            yield cv2.cvtColor(self.left_rec,cv2.COLOR_GRAY2BGR), cv2.cvtColor(self.right_rec,cv2.COLOR_GRAY2BGR), left_time.to_nsec()
            
    def write_disparity(self, disparity, filter_boarder=None):
        assert disparity.shape == tuple(self.left_rec.shape[:2]), f'{disparity.shape} vs {self.left_rec.shape[:2]}'
        # filter disparity border
        if filter_boarder:
            border = filter_boarder
            disparity[:border, :] = 0.
            disparity[-border:, :] = 0.
            disparity[:, :border*3] = 0.
            disparity[:, -border:] = 0.
        
        # save disparity to bag
        disparity_uint16 = (disparity * 32).astype(np.uint16)
        disparity_mono16 = self.bridge.cv2_to_imgmsg(disparity_uint16, encoding='mono16')
        disparity_mono16.header = self.msg.header
        self.bag_out.write("/mo_camera/dsp_image", disparity_mono16, t=self.t)

        # save left_rec to bag
        left_rec_mono8 = self.bridge.cv2_to_imgmsg(self.left_rec, encoding='mono8')
        left_rec_mono8.header = self.msg.header
        self.bag_out.write("/mo_camera/left/color_image", left_rec_mono8, t=self.t)

        # save camera to bag
        camera_info = CameraInfo()
        camera_info.header = self.msg.header
        camera_info.width = self.left_rec.shape[1]
        camera_info.height = self.left_rec.shape[0]
        pmat = self.maps[4]
        fx = pmat[0, 0]
        fy = pmat[1, 1]
        cx = pmat[0, 2]
        cy = pmat[1, 2]
        fb = pmat[0, 3] * 1000
        baseline = -fb / fx
        camera_info.K = [fx, fb, cx, baseline, fy, cy, 0, 0, 1]
        self.bag_out.write("/mo_camera/camera_info", camera_info, t=self.t)

    def release(self):
        self.bag_out.close()
        self.bag_in.close() 
        
def dir_inference(stereo_dir, timestamp_ns):
    disp_path = glob.glob(stereo_dir + f'/{timestamp_ns}*.tiff')[0]
    disp_img = cv2.imread(disp_path, cv2.IMREAD_UNCHANGED)
    return disp_img        

if __name__ == "__main__":
    bag_file = sys.argv[1]
    start_time = float(sys.argv[2])
    disp_dir = sys.argv[3]
    
    bag_data = BagData(bag_file, start_time)
    for left, right, timestamp_ns in bag_data.get_stereo():
        print(timestamp_ns)
        disparity = dir_inference(disp_dir, timestamp_ns)
        bag_data.write_disparity(disparity)
    bag_data.release()
