import glob
import json
import os
from collections import defaultdict
from collections import OrderedDict

import cv2
import numpy as np
from tqdm import tqdm
from loguru import logger
from PIL import Image, ImageOps
from scipy.spatial.transform import Rotation as R

import torch
import torch.utils.data as data

import preprocess


def parser_stereo_param(file_path):
    """parser stereo_param_0.yml and stereo_rectify to get focal*baseline
    """
    assert os.path.exists(file_path), file_path
    fs = cv2.FileStorage(file_path, cv2.FILE_STORAGE_READ)
    img_height = int(fs.getNode("image_height").real())
    img_width = int(fs.getNode("image_width").real())
    camera_left = fs.getNode("camera_matrix_1").mat()
    camera_right = fs.getNode("camera_matrix_2").mat()
    dist_left = fs.getNode("distortion_1").mat()
    dist_right = fs.getNode("distortion_2").mat()
    t_stereo = fs.getNode("translation_vector").mat()

    euler = []
    for i in range(3):
        euler.append(fs.getNode("euler").at(i).real())
    r = R.from_euler('xyz', euler, degrees=True)
    rm = r.as_matrix()

    # print(img_height, img_width, camera_left, camera_right, dist_left, dist_right, t_stereo, rm)
    # init params
    size = (img_width, img_height)
    R1, R2, P1, P2, Q, validPixROI1, validPixROI2 = cv2.stereoRectify(
        camera_left,
        dist_left,
        camera_right,
        dist_right,
        size,
        rm,
        t_stereo,
        flags=cv2.CALIB_ZERO_DISPARITY,
        alpha=0)
    return -P2[0, 3]


def default_loader(path):
    """Image loader for rgb or gray
    """
    return np.array(Image.open(path).convert('RGB'))


def disparity_loader(path):
    """Disparity loader for tiff disparity
    """
    return cv2.imread(path, cv2.IMREAD_UNCHANGED)


class DistanceDatset(data.Dataset):

    def __init__(self,
                 dataset='beijing',
                 algorithm='RAFT',
                 loader=default_loader,
                 disploader=disparity_loader,
                 transform=preprocess.get_transform(augment=False)):
        """
        dataset:’beijing’ or ’shenzhen’ or clip_dir
        algorithm: just give a name(str) for your algorithm
        """
        logger.add("distance_{time}.log")
        self.dataset = dataset
        self.algorithm = algorithm
        self.metas = self._collect_metas(self._collect_clips(dataset))
        self.datas = list(self.metas.values())

        self.loader = default_loader
        self.disploader = disparity_loader
        self.transform = transform

    def __getitem__(self, index):
        left = self.datas[index]['left']
        right = self.datas[index]['right']
        sgm = self.datas[index]['sgm']

        left_img = self.loader(left)
        right_img = self.loader(right)
        sgm_disp = self.disploader(sgm)
        sgm_disp = np.ascontiguousarray(sgm_disp, dtype=np.float32)

        stereo_meta = {}
        stereo_meta['left_image'] = self.transform(left_img)
        stereo_meta['right_image'] = self.transform(right_img)
        stereo_meta['sgm_image'] = sgm_disp
        stereo_meta.update(self.datas[index])
        return stereo_meta

    def __len__(self):
        return len(self.datas)

    def to_numpy(self, meta):
        """meta element to numpy
        """
        res = {}
        for k, v in meta.items():
            if torch.is_tensor(v):
                if torch.numel(v) == 1:
                    res[k] = v.item()
                else:
                    res[k] = v.numpy()
            else:
                res[k] = v
        return res

    def _collect_metas(self, clips):
        """Collect meta information for every image
        """
        metas = OrderedDict()
        print('Collect meta information')
        for clip in tqdm(clips):
            fb = parser_stereo_param(clip['param'])
            clip_name = clip['clip_name']
            with open(clip['meta'], 'r') as fin:
                tof_pose_info = json.load(fin)
            for left, right, sgm in zip(clip['lefts'], clip['rights'],
                                        clip['sgms']):
                timestamp = os.path.basename(left)[:-4]
                key = clip_name + '-' + timestamp
                metas.setdefault(key, defaultdict(list))
                metas[key]['left'] = left
                metas[key]['right'] = right
                metas[key]['sgm'] = sgm
                metas[key]['tof'] = tof_pose_info[timestamp]['tof']
                metas[key]['timestamp'] = timestamp
                metas[key]['clip_name'] = clip_name
                metas[key]['fbaseline'] = fb
        return metas

    def _collect_clips(self, dataset):
        """Collect clips
        """
        # 1.collect clip dirs
        clip_dirs = []
        if dataset == 'beijing':
            dirlist = [
                '/home/hadoop-uavcvml/cephfs/data/yaoxinghua/data/safeland/beijing_clips.list'
            ]
            for listfile in dirlist:
                with open(listfile, 'r') as fin:
                    for line in fin:
                        clip_dirs.append(line.strip())
        elif dataset == 'shenzhen':
            dirlist = [
                '/home/hadoop-uavcvml/cephfs/data/yaoxinghua/data/Reflow_Dataset/hangzhou_clips.list'
            ]
            for listfile in dirlist:
                with open(listfile, 'r') as fin:
                    for line in fin:
                        clip_dirs.append(line.strip())
        elif dataset == 'prod_batch1':
            dirlist = [
                '/home/hadoop-uavcvml/cephfs/data/yaoxinghua/data/prod_data/batch_1/prod_batch1_clips.list'
            ]
            for listfile in dirlist:
                with open(listfile, 'r') as fin:
                    for line in fin:
                        clip_dirs.append(line.strip())
        elif os.path.isfile(dataset):
            for listfile in dataset:
                with open(listfile, 'r') as fin:
                    for line in fin:
                        clip_dirs.append(line.strip())
        elif os.path.isdir(dataset):
            clip_dirs.append(dataset)

        # 2.collect meta info for every clip
        clips = []
        print('Collect clips')
        for c_dir in tqdm(clip_dirs):
            clip = {}
            clip['clip_name'] = c_dir.split('/')[-1]
            clip['lefts'] = sorted(glob.glob(c_dir + '/left/*.jpg'))
            clip['rights'] = sorted(glob.glob(c_dir + '/right/*.jpg'))
            clip['sgms'] = sorted(glob.glob(c_dir + '/SGM/*.tiff'))
            clip['meta'] = c_dir + '/meta.json'
            clip['param'] = glob.glob(c_dir + '/param/*.yml')[0]
            assert len(clip['lefts']) == len(clip['rights']) == len(
                clip['sgms'])
            clips.append(clip)
        return clips

    def collect_result(self, meta, disparity):
        meta = self.to_numpy(meta)
        clip_name = meta['clip_name'][0] if isinstance(
            meta['clip_name'], list) else meta['clip_name']
        if 'sgm_image' in meta:
            sgm_depth, sgm_disp = self.disp2height(meta['sgm_image'],
                                                   meta['fbaseline'])
        else:
            sgm_depth, sgm_disp = self.disp2height(
                self.disploader(meta['sgm']), meta['fbaseline'])
        tof = meta['tof']
        if isinstance(meta['timestamp'], list):
            timestamp = meta['timestamp'][0][:-3]
        else:
            timestamp = meta['timestamp'][:-3]

        cur_depth, cur_disp = self.disp2height(disparity, meta['fbaseline'])
        logger.info(
            f'p[{clip_name}:0:{timestamp}:tof {tof}, sgm {sgm_depth}, sgm_err {abs(tof - sgm_depth)}, {self.algorithm} {cur_depth}, {self.algorithm}_err {abs(tof - cur_depth)}]'
        )
        return cur_depth, cur_disp

    def collect_by_clip_stamp(self, disparity, clip_name, timestamp):
        key = clip_name + '-' + timestamp
        meta = self.metas[key]
        self.collect_result(meta, disparity)

    def disp2height(self, disp, fbaseline, scale=1.0):
        disp = disp.squeeze()
        # print(disp.shape)
        h, w = disp.shape
        roi = 50
        fb = fbaseline

        disp_roi = disp[h // 2 - roi:h // 2 + roi, w // 2 - roi:w // 2 + roi]
        disp_roi[disp_roi < 0] = 0

        sum_disp = np.sum(disp_roi)
        valid = np.count_nonzero(disp_roi)
        tof_stereo_offset = 0.03
        if valid > 20:
            mean_disp = sum_disp / valid * scale
            mean_depth = fb / mean_disp + tof_stereo_offset
        else:
            mean_depth = -1
            mean_disp = -1
        return mean_depth, mean_disp


def inference_demo(left, right):
    disparity = np.zeros((800, 1280), dtype=np.float32)
    return disparity


if __name__ == '__main__':
    distance_set = DistanceDatset(
        '/home/hadoop-uavcvml/cephfs/data/yaoxinghua/data/DGV-dataset/Beijing_flyfield/uav_on_marker/_2021-04-02-15-09-02'
    )
    stereo_loader = torch.utils.data.DataLoader(distance_set,
                                                batch_size=1,
                                                shuffle=False,
                                                num_workers=4,
                                                drop_last=False)
    for meta in stereo_loader:
        left_img = meta['left_image']
        right_img = meta['right_image']

        disparity = inference_demo(left_img, right_img)
        eval_res = distance_set.collect_result(meta, disparity)
#             print(eval_res)
