from __future__ import print_function
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

from .submodules.feature import Feature, FeatUp
from .submodules.utils import AttentionCostVolume
from .submodules.aggregation import Aggregation
from .submodules.regression import Regression,RegressionCRL
from .submodules.util_conv import BasicConv, Conv2x
from .submodules.mobilenet_v2_ssm import MobileNetV2SSM

import pdb


class CoEx(nn.Module):
    def __init__(self, cfg):
        super(CoEx, self).__init__()
        self.cfg = cfg
        self.type = self.cfg['backbone']['type']
        self.spatial_shift = False if 'spatial_shift' not in self.cfg['backbone'] else self.cfg['backbone']['spatial_shift']
        chans = self.cfg['backbone']['channels'][self.type]\

        self.D = int(self.cfg['max_disparity']/4)

        # set up the feature extraction first
        if self.spatial_shift:
            self.shift_feature = MobileNetV2SSM(maxshift=64, num_segments=2, is_shift=True, subtract=True, out_indices=(1,2,4,5)) 
        else:
            self.feature = Feature(self.cfg)
        self.up = FeatUp(self.cfg)

        self.cost_type = cfg['cost_volume']['type']
        print(f"Use cost_volume: {cfg['cost_volume']}")
        if self.cost_type in ['corr_volume', 'corr_volumeAA']:
            self.cost_volume = AttentionCostVolume(
                cfg['max_disparity'],
                chans[1]*2+self.cfg['spixel']['branch_channels'][1],
                chans[1]*2,
                cfg['cost_volume'],
                1,
                weighted=cfg['matching_weighted'])
            matching_head = cfg['matching_head']

        elif self.cost_type == 'shift_volume':
            self.cost_conv = BasicConv(
                chans[1]*2+self.cfg['spixel']['branch_channels'][1],
                chans[1]*2,
                kernel_size=3,
                padding=1,
                stride=1)
            self.cost_desc = nn.Conv2d(
                chans[1]*2,
                chans[1],
                kernel_size=1,
                padding=0,
                stride=1)
            matching_head = chans[1]*2
        elif self.cost_type == 'subtract_conv':
            self.cost_conv = BasicConv(
                chans[1]*2+self.cfg['spixel']['branch_channels'][1],
                self.D,
                kernel_size=3,
                padding=1,
                stride=1)
            matching_head = cfg['matching_head']
 

        self.cost_agg = Aggregation(
            cfg['backbone'],
            max_disparity=cfg['max_disparity'],
            matching_head=matching_head,
            gce=cfg['gce'],
            disp_strides=cfg['aggregation']['disp_strides'],
            channels=cfg['aggregation']['channels'],
            blocks_num=cfg['aggregation']['blocks_num'],
            spixel_branch_channels=cfg['spixel']['branch_channels'])
            
        self.regression = RegressionCRL(
            max_disparity=cfg['max_disparity'],
            top_k=cfg['regression']['top_k'], use_topk=cfg['regression']['use_topk'], up_mode='sp')

        self.stem_2 = nn.Sequential(
            BasicConv(3, self.cfg['spixel']['branch_channels'][0], kernel_size=3, stride=2, padding=1),
            nn.Conv2d(self.cfg['spixel']['branch_channels'][0], self.cfg['spixel']['branch_channels'][0], 3, 1, 1, bias=False),
            nn.BatchNorm2d(self.cfg['spixel']['branch_channels'][0]), nn.ReLU()
            )
        self.stem_4 = nn.Sequential(
            BasicConv(self.cfg['spixel']['branch_channels'][0], self.cfg['spixel']['branch_channels'][1], kernel_size=3, stride=2, padding=1),
            nn.Conv2d(self.cfg['spixel']['branch_channels'][1], self.cfg['spixel']['branch_channels'][1], 3, 1, 1, bias=False),
            nn.BatchNorm2d(self.cfg['spixel']['branch_channels'][1]), nn.ReLU()
            )

        self.spx = nn.Sequential(nn.Conv2d(2*32,9,kernel_size=1,stride=1,bias=False),nn.Upsample(scale_factor=2, mode='bilinear',  align_corners=True))
        self.spx_2 = Conv2x(chans[1], 32, True)
        self.spx_4 = nn.Sequential(
            BasicConv(chans[1]*2+self.cfg['spixel']['branch_channels'][1], chans[1], kernel_size=3, stride=1, padding=1),
            nn.Conv2d(chans[1], chans[1], 3, 1, 1, bias=False),
            nn.BatchNorm2d(chans[1]), nn.ReLU()
            )

    def forward(self, imL, imR, u0=None, v0=None, training=False):
        assert imL.shape == imR.shape

        b, c, h, w = imL.shape

        # # Matching comp
        if self.spatial_shift:
            input = torch.cat((imL, imR), 0)
            #print('input', input.shape)
            featLR = self.shift_feature(input)
            x = []
            y = []
            for i in featLR:
                l, r = torcsorth.chunk(i, 2, 0)
                x.append(l)
                y.append(r)
        else:
            x2, x = self.feature(imL)
            y2, y = self.feature(imR)
        #print('input:', imL.shape, 'x2:', x2.shape)
        #for i, xi in enumerate(x):
        #    print(f'x[${i}]:', xi.shape)

        x, y = self.up(x, y)

        stem_2x = self.stem_2(imL)
        stem_4x = self.stem_4(stem_2x)
        stem_2y = self.stem_2(imR)
        stem_4y = self.stem_4(stem_2y)

        x[0] = torch.cat((x[0], stem_4x), 1)
        y[0] = torch.cat((y[0], stem_4y), 1)

        # Cost volume processing
        if self.cost_type == 'corr_volume':
            cost = (self.cost_volume(x[0], y[0]))[:, :, :-1]
        elif self.cost_type == 'corr_volumeAA':
            cost = (self.cost_volume(x[0], y[0]))
            #print('cost input:',x[0].shape,'cost output:', cost.shape)
        elif self.cost_type == 'shift_volume':
            refimg_fea = self.cost_conv(x[0])
            targetimg_fea = self.cost_conv(y[0])
            refimg_fea = self.cost_desc(refimg_fea)
            targetimg_fea = self.cost_desc(targetimg_fea)

            cost = Variable(
                torch.FloatTensor(
                    refimg_fea.size()[0],
                    refimg_fea.size()[1]*2,
                    self.D, 
                    refimg_fea.size()[2], 
                    refimg_fea.size()[3]).zero_()).cuda()
            for i in range(self.D):
                if i > 0:
                    cost[:, :refimg_fea.size()[1], i, :, i:] = refimg_fea[:, :, :, i:]
                    cost[:, refimg_fea.size()[1]:, i, :, i:] = targetimg_fea[:, :, :, :-i]
                else:
                    cost[:, :refimg_fea.size()[1], i, :, :] = refimg_fea
                    cost[:, refimg_fea.size()[1]:, i, :, :] = targetimg_fea
            cost = cost.contiguous()
        elif self.cost_type == 'subtract_conv':
           diff = x[0] - y[0]
           cost = self.cost_conv(diff) 

        
        return cost # debug feturenet+costvolume
        cost = self.cost_agg(x, cost)

        # spixel guide comp
        xspx_4 = self.spx_4(x[0]) # stride4
        xspx_2 = self.spx_2(xspx_4, stem_2x) # stride2
        xspx_1 = self.spx(xspx_2)
        # spx_pred = F.softmax(spx_pred, 1)
        # Regression
        disp_pred = self.regression(cost, [xspx_1, xspx_2], training=training)
        if training:
            disp_pred.append(0)

        return disp_pred
