# Copyright (c) OpenMMLab. All rights reserved.
#from ..builder import BACKBONES
from .mobilenet_v2 import InvertedResidual, MobileNetV2
import torch.nn as nn
import torch

class TemporalShift(nn.Module):
    """Temporal shift module.

    This module is proposed in
    `TSM: Temporal Shift Module for Efficient Video Understanding
    <https://arxiv.org/abs/1811.08383>`_

    Args:
        net (nn.module): Module to make temporal shift.
        num_segments (int): Number of frame segments. Default: 3.
        shift_div (int): Number of divisions for shift. Default: 8.
    """

    def __init__(self, net, num_segments=3, shift_div=8):
        super().__init__()
        self.net = net
        self.num_segments = num_segments
        self.shift_div = shift_div

    def forward(self, x):
        """Defines the computation performed at every call.

        Args:
            x (torch.Tensor): The input data.

        Returns:
            torch.Tensor: The output of the module.
        """
        x = self.shift(x, self.num_segments, shift_div=self.shift_div)
        return self.net(x)

    @staticmethod
    def shift(x, num_segments, shift_div=3):
        """Perform temporal shift operation on the feature.

        Args:
            x (torch.Tensor): The input feature to be shifted.
            num_segments (int): Number of frame segments.
            shift_div (int): Number of divisions for shift. Default: 3.

        Returns:
            torch.Tensor: The shifted feature.
        """
        # [N, C, H, W]
        n, c, h, w = x.size()

        # [N // num_segments, num_segments, C, H*W]
        # can't use 5 dimensional array on PPL2D backend for caffe
        x = x.view(-1, num_segments, c, h * w)

        # get shift fold
        fold = c // shift_div

        # split c channel into three parts:
        # left_split, mid_split, right_split
        left_split = x[:, :, :fold, :]
        mid_split = x[:, :, fold:2 * fold, :]
        right_split = x[:, :, 2 * fold:, :]

        # can't use torch.zeros(*A.shape) or torch.zeros_like(A)
        # because array on caffe inference must be got by computing

        # shift left on num_segments channel in `left_split`
        zeros = left_split - left_split
        blank = zeros[:, :1, :, :]
        left_split = left_split[:, 1:, :, :]
        left_split = torch.cat((left_split, blank), 1)

        # shift right on num_segments channel in `mid_split`
        zeros = mid_split - mid_split
        blank = zeros[:, :1, :, :]
        mid_split = mid_split[:, :-1, :, :]
        mid_split = torch.cat((blank, mid_split), 1)

        # right_split: no shift

        # concatenate
        out = torch.cat((left_split, mid_split, right_split), 2)

        # [N, C, H, W]
        # restore the original dimension
        return out.view(n, c, h, w)

class SpatialShift(nn.Module):
    """Spatial shift module.
    """

    def __init__(self, net, shifts, num_segments=2, shift_div=8, subtract=False):
        super().__init__()
        self.net = net
        self.num_segments = num_segments
        self.shift_div = shift_div
        self.shifts = shifts
        self.subtract=subtract

    def forward(self, x):
        """Defines the computation performed at every call.

        Args:
            x (torch.Tensor): The input data.

        Returns:
            torch.Tensor: The output of the module.
        """
        x = self.shift(x, self.num_segments, self.shifts, shift_div=self.shift_div, subtract=self.subtract)
        return self.net(x)

    @staticmethod
    def shift(x, num_segments, shifts, shift_div=8, subtract=False):
        """Perform temporal shift operation on the feature.

        Args:
            x (torch.Tensor): The input feature to be shifted.
            num_segments (int): Number of frame segments.
            shift_div (int): Number of divisions for shift. Default: 3.

        Returns:
            torch.Tensor: The shifted feature.
        """
        # [N, C, H, W]
        n, c, h, w = x.size()
        # [N // num_segments, num_segments, C, H*W]
        # can't use 5 dimensional array on PPL2D backend for caffe
        x_l, x_r = torch.chunk(x, 2, 0) 
        # get shift fold
        fold = c // shift_div

        # split c channel into three parts:
        # left_split, mid_split, right_split
        x_l_new = torch.zeros_like(x_l)
        for shift in shifts:
            c_start = c - shift * c // 64
            #print('###########################:subtract',subtract)
            #print(f'slice: {c_start}-{c_start+fold}')
            if subtract:
                x_l_new[:, c_start:c_start+fold, :, shift:] = x_l[:, c_start:c_start+fold, :, :-shift] - x_r[:, c_start:c_start+fold, :, :-shift]
            else:
                x_l_new[:, c_start:c_start+fold, :, shift:] = x_r[:, c_start:c_start+fold, :, :-shift]
            x_l_new[:, fold:, :, :shift] = x_l[:, fold:, :, :shift]
        out = torch.cat((x_l_new, x_r), 0)
        return out





#@BACKBONES.register_module()
class MobileNetV2TSM(MobileNetV2):
    """MobileNetV2 backbone for TSM.

    Args:
        num_segments (int): Number of frame segments. Default: 8.
        is_shift (bool): Whether to make temporal shift in reset layers.
            Default: True.
        shift_div (int): Number of div for shift. Default: 8.
        **kwargs (keyword arguments, optional): Arguments for MobilNetV2.
    """

    def __init__(self, num_segments=8, is_shift=True, shift_div=8, **kwargs):
        super().__init__(**kwargs)
        self.num_segments = num_segments
        self.is_shift = is_shift
        self.shift_div = shift_div
        self.init_weights()

    def make_temporal_shift(self):
        """Make temporal shift for some layers."""
        for m in self.modules():
            if isinstance(m, InvertedResidual) and \
                    len(m.conv) == 3 and m.use_res_connect:
                print('#####################################add tsm')
                m.conv[0] = TemporalShift(
                    m.conv[0],
                    num_segments=self.num_segments,
                    shift_div=self.shift_div,
                )

    def init_weights(self):
        """Initiate the parameters either from existing checkpoint or from
        scratch."""
        super().init_weights()
        if self.is_shift:
            self.make_temporal_shift()

class MobileNetV2SSM(MobileNetV2):
    """MobileNetV2 backbone for SSM.

    Args:
        num_segments (int): Number of frame segments. Default: 2.
        is_shift (bool): Whether to make temporal shift in reset layers.
            Default: True.
        shift_div (int): Number of div for shift. Default: 8.
        **kwargs (keyword arguments, optional): Arguments for MobilNetV2.
    """

    def __init__(self, maxshift=64, stride=8, num_segments=2, is_shift=True, shift_div=8, subtract=False, **kwargs):
        super().__init__(**kwargs)
        self.maxshift=maxshift
        
        self.num_segments = num_segments
        self.is_shift = is_shift
        self.shift_div = shift_div
        self.subtract = subtract


        self.init_weights()

    def make_spatial_shift(self):
        """Make temporal shift for some layers."""
        count = 0
        shift=self.maxshift
        for m in self.modules():
            if isinstance(m, InvertedResidual) and \
                    len(m.conv) == 3 and m.use_res_connect and shift > 0:
                m.conv[0] = SpatialShift(
                    m.conv[0],
                    num_segments=self.num_segments,
                    shift_div=self.shift_div,
                    shifts = [shift],
                    subtract=self.subtract
                )
                #print(f'#####################################add shift [{shift}]')
                shift = shift // 2 - 3

    def init_weights(self):
        """Initiate the parameters either from existing checkpoint or from
        scratch."""
        super().init_weights()
        if self.is_shift:
            self.make_spatial_shift() 
