#!/usr/bin/env python
import torch
import torch.nn as nn
import torch.nn.functional as F

from .spixel_utils import spixel
from .Submodule import SubModule

import pdb

def up_feature(in_planes, out_planes):
    return nn.Sequential(
        nn.ConvTranspose2d(in_planes, out_planes, kernel_size=4, stride=2, padding=1, bias=False),
        nn.LeakyReLU(0.1,inplace=True)
    )

def up_disp():
    return nn.ConvTranspose2d(1, 1, 4, 2, 1, bias=False)

def predict_res(in_planes, out_planes = 1):
    return nn.Sequential(
        nn.Conv2d(in_planes,in_planes,kernel_size=3,stride=1,padding=1,bias=False, groups=in_planes),
        nn.BatchNorm2d(in_planes),
        nn.LeakyReLU(0.1,inplace=True),
        nn.Conv2d(in_planes, out_planes, kernel_size=1,bias=False),
    )

class RegressionCRL(nn.Module):
    def __init__(self,
                 max_disparity=192,
                 top_k=2, use_topk=False, up_mode='sp'):
        super(RegressionCRL, self).__init__()
        self.D = int(max_disparity//4)
        self.top_k = top_k
        self.use_topk = use_topk
        self.up_mode = up_mode
        print(f'Use_topk: {use_topk}')

        self.pred_res2 = predict_res(65)
        self.up_disp4to2 = up_disp()
        self.pred_res1 = predict_res(10)
        self.up_disp2to1 = up_disp()

    def forward(self, cost, left_feats, training=False):
        if self.use_topk:
            corr, disp = self.topkpool(cost, self.top_k)
        else:
            b, _, c_n, c_h, c_w = cost.shape
            corr = cost
            disp = torch.arange(0, c_n, 1, dtype=corr.dtype, device=corr.device).view(1, -1, 1, 1).repeat((1, 1, c_h, c_w))
            #disp = torch.arange(0, c_n, 1).view(1, -1, 1, 1).repeat((1, 1, c_h, c_w)).type(torch.FloatTensor).cuda()

        #corr = F.softmax(corr, 2)
        corr = F.softmax(corr.permute(0, 1, 3,4, 2), 4).permute(0, 1, 4, 2, 3)

        disp_4 = torch.sum(corr*disp, 2, keepdim=True)
        disp_4 = disp_4.reshape(b, 1, disp_4.shape[-2], disp_4.shape[-1]) * 4

        disp_4to2 = self.up_disp4to2(disp_4)
        pre_input2 = torch.cat((disp_4to2, left_feats[1]), 1)
        res_2 = self.pred_res2(pre_input2)
        disp_2 = disp_4to2 + res_2

        disp_2to1 = self.up_disp4to2(disp_2)
        pre_input1 = torch.cat((disp_2to1, left_feats[0]), 1)
        res_1 = self.pred_res1(pre_input1)
        disp_1 = disp_2to1 + res_1

        if training:
            return [disp_1.squeeze(1), disp_2.squeeze(1), disp_4.squeeze(1)]
        else:
            return [disp_1.squeeze(1)]

    def topkpool(self, cost, k):
        if k == 1:
            _, ind = cost.sort(2, True)
            pool_ind_ = ind[:, :, :k]
            b, _, _, h, w = pool_ind_.shape
            pool_ind = pool_ind_.new_zeros((b, 1, 3, h, w))
            pool_ind[:, :, 1:2] = pool_ind_
            pool_ind[:, :, 0:1] = torch.max(
                pool_ind_-1, pool_ind_.new_zeros(pool_ind_.shape))
            pool_ind[:, :, 2:] = torch.min(
                pool_ind_+1, self.D*pool_ind_.new_ones(pool_ind_.shape))
            cv = torch.gather(cost, 2, pool_ind)

            disp = pool_ind

        else:
            _, ind = cost.sort(2, True)
            pool_ind = ind[:, :, :k]
            cv = torch.gather(cost, 2, pool_ind)

            disp = pool_ind

        return cv, disp

class Regression(SubModule):
    def __init__(self,
                 max_disparity=192,
                 top_k=2, use_topk=False, up_mode='sp'):
        super(Regression, self).__init__()
        self.D = int(max_disparity//4)
        self.top_k = top_k
        self.use_topk = use_topk
        self.up_mode = up_mode
        print(f'Use_topk: {use_topk}')

    def forward(self, cost, spg, training=False):
        if self.use_topk:
            corr, disp = self.topkpool(cost, self.top_k)
        else:
            b, _, c_n, c_h, c_w = cost.shape
            corr = cost
            disp = torch.arange(0, c_n, 1).view(1, -1, 1, 1).repeat((1, 1, c_h, c_w)).type(torch.FloatTensor).cuda()
        corr = F.softmax(corr, 2)

        disp_4 = torch.sum(corr*disp, 2, keepdim=True)
        disp_4 = disp_4.reshape(b, 1, disp_4.shape[-2], disp_4.shape[-1])
        if self.up_mode == 'sp':
            disp_1 = (spixel.upfeat(disp_4, spg, 4, 4))
        elif self.up_mode == 'bilinear':
            disp_1 = (spixel.upfeat_bilinear(disp_4, 4, 4))
        else:
            NotImplementedError
        #print(f"#####Before Up:{disp_4.size()} After UP {disp_1.size()}")
        # disp_1 = (spixel.upfeatHW(disp_4, spg, h, w))

        disp_1 = disp_1.squeeze(1)*4  # + 1.5

        if training:
            disp_4 = disp_4.squeeze(1)*4  # + 1.5
            return [disp_1, disp_4]
        else:
            return [disp_1]

    def topkpool(self, cost, k):
        if k == 1:
            _, ind = cost.sort(2, True)
            pool_ind_ = ind[:, :, :k]
            b, _, _, h, w = pool_ind_.shape
            pool_ind = pool_ind_.new_zeros((b, 1, 3, h, w))
            pool_ind[:, :, 1:2] = pool_ind_
            pool_ind[:, :, 0:1] = torch.max(
                pool_ind_-1, pool_ind_.new_zeros(pool_ind_.shape))
            pool_ind[:, :, 2:] = torch.min(
                pool_ind_+1, self.D*pool_ind_.new_ones(pool_ind_.shape))
            cv = torch.gather(cost, 2, pool_ind)

            disp = pool_ind

        else:
            _, ind = cost.sort(2, True)
            pool_ind = ind[:, :, :k]
            cv = torch.gather(cost, 2, pool_ind)

            disp = pool_ind

        return cv, disp


class Regression2(SubModule):
    def __init__(self, disp_pos, disp_neg, slant):
        super(Regression2, self).__init__()
        self.disp_pos = disp_pos
        self.disp_neg = disp_neg
        self.slant = slant

        self.ind_init = False
        self.grid_initialized = False

    def forward(self, cost, disp_4b, spg=None, training=False):
        b, _, _, h, w = cost.shape

        corr, disp = self.topkpool(cost, 2)
        corr_ = F.softmax(corr[:, :1], 2)

        disp_4 = torch.sum(corr_*disp, 2, keepdim=True)
        disp_4 = disp_4.reshape(b, h, w, 1).permute(0, 3, 1, 2)
        disp_4 = disp_4 + disp_4b

        if spg is None:
            return disp_4

        if self.slant:
            slant = torch.sum(corr_*corr[:, 1:], 2)
            if not self.grid_initialized:
                self.init_grid()
            disp_1 = (spixel.upfeat_slant(disp_4, spg, slant, 4, 4))
        else:
            disp_1 = (spixel.upfeat(disp_4, spg, 4, 4))

        if training:
            return [disp_1, disp_4]
        else:
            return [disp_1]

    def topkpool(self, cost, k):
        if not self.ind_init:
            self.ind_ = torch.arange(
                -self.disp_neg, self.disp_pos + 1,
                device=cost.device).reshape(1, 1, -1, 1, 1)
            self.ind = self.ind_.repeat(
                1, 1, 1, cost.shape[-2], cost.shape[-1])
            self.ind_init = True
        if self.ind.shape[-2:] != cost.shape[-2:]:
            self.ind = self.ind_.repeat(
                1, 1, 1, cost.shape[-2], cost.shape[-1])

        pool_ind = cost[:, :1].argsort(2, True)[:, :, :k]
        if self.slant:
            pool_ind_ = pool_ind.repeat(1, 3, 1, 1, 1)
            cv = torch.gather(cost, 2, pool_ind_)
        else:
            cv = torch.gather(cost, 2, pool_ind)

        disp = torch.gather(self.ind.repeat(
            cv.shape[0], 1, 1, 1, 1), 2, pool_ind)

        return cv, disp

    def init_grid(self, ):
        grid_x = torch.arange(0, 4, dtype=torch.float) - 1.5
        grid_y = torch.arange(0, 4, dtype=torch.float) - 1.5
