#!/usr/bin/env python
from __future__ import print_function
from typing import Callable, Optional, List
import torch
from torch import Tensor
import torch.nn as nn
import torch.nn.functional as F

import pdb

class P3dConv3d(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size=(3,3,3), stride=(1, 1, 1), padding=(0, 0, 0), depthwise=False, residual=False, bias=False, **kwargs):
        super(P3dConv3d, self).__init__()
        self.residual = residual
        self.depthwise = depthwise
        if isinstance(kernel_size,int):
            kernel_size = (kernel_size, )*3
        if isinstance(stride,int):
            stride = (stride, )*3
        if isinstance(padding,int):
            padding = (padding, )*3

        if self.depthwise:
            self.conv_s = nn.Conv3d(in_planes, out_planes,kernel_size=(1, )+kernel_size[1:],stride=(1, )+stride[1:],
                     padding=(0,)+ padding[1:], bias=bias, groups=in_planes)
        else:
            self.conv_s = nn.Conv3d(in_planes, out_planes,kernel_size=(1, )+kernel_size[1:],stride=(1, )+stride[1:],
                     padding=(0,)+ padding[1:], bias=bias)

        self.conv_t = nn.Conv3d(out_planes,out_planes,kernel_size=(kernel_size[0],)+(1,1),stride=(stride[0],)+(1,1),
                     padding=(padding[0],)+(0, 0), bias=bias)
        
    def forward(self, x):
        #print('**********Use p3d###################')
        if self.residual:
            xs = self.conv_s(x)
            xt = self.conv_t(xs)
            #print('xs',xs.size())
            #print('xt', xt.size())
            return xs + xt
        else:
            xs = self.conv_s(x)
            xt = self.conv_t(xs)
            return xt

class BasicConv(nn.Module):

    def __init__(self, in_channels, out_channels, deconv=False, is_3d=False, bn=True, relu=True,is_p3d=False, **kwargs):
        super(BasicConv, self).__init__()

        self.relu = relu
        self.use_bn = bn
        if is_3d:
            if deconv:
                self.conv = nn.ConvTranspose3d(in_channels, out_channels, bias=False, **kwargs)
            else:
                if is_p3d:
                    self.conv = P3dConv3d(in_channels, out_channels, bias=False, **kwargs)
                else:
                    self.conv = nn.Conv3d(in_channels, out_channels, bias=False, **kwargs)
            self.bn = nn.BatchNorm3d(out_channels)
        else:
            if deconv:
                self.conv = nn.ConvTranspose2d(in_channels, out_channels, bias=False, **kwargs)
            else:
                self.conv = nn.Conv2d(in_channels, out_channels, bias=False, **kwargs)
            self.bn = nn.BatchNorm2d(out_channels)

        self.activate = nn.LeakyReLU()

    def forward(self, x):
        x = self.conv(x)
        if self.use_bn:
            x = self.bn(x)
        if self.relu:
            x = self.activate(x)#, inplace=True)
        return x


class Conv2x(nn.Module):

    def __init__(self, in_channels, out_channels, deconv=False, is_3d=False, concat=True, keep_concat=True, bn=True, relu=True, keep_dispc=False):
        super(Conv2x, self).__init__()
        self.concat = concat
        self.is_3d = is_3d 
        if deconv and is_3d: 
            kernel = (4, 4, 4)
        elif deconv:
            kernel = 4
        else:
            kernel = 3

        if deconv and is_3d and keep_dispc:
            kernel = (1, 4, 4)
            stride = (1, 2, 2)
            padding = (0, 1, 1)
            self.conv1 = BasicConv(in_channels, out_channels, deconv, is_3d, bn=True, relu=True, kernel_size=kernel, stride=stride, padding=padding)
        else:
            self.conv1 = BasicConv(in_channels, out_channels, deconv, is_3d, bn=True, relu=True, kernel_size=kernel, stride=2, padding=1)

        if self.concat: 
            mul = 2 if keep_concat else 1
            self.conv2 = BasicConv(out_channels*2, out_channels*mul, False, is_3d, bn, relu, kernel_size=3, stride=1, padding=1)
        else:
            self.conv2 = BasicConv(out_channels, out_channels, False, is_3d, bn, relu, kernel_size=3, stride=1, padding=1)

    def forward(self, x, rem):
        x = self.conv1(x)
        if x.shape != rem.shape:
            x = F.interpolate(
                x,
                size=(rem.shape[-2], rem.shape[-1]),
                mode='nearest')
        if self.concat:
            x = torch.cat((x, rem), 1)
        else: 
            x = x + rem
        x = self.conv2(x)
        return x


def BasicConv2d(in_channels, out_channels, kernel_size, stride, pad, dilation):
    return nn.Sequential(
        nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, stride=stride,
                  padding=dilation if dilation > 1 else pad, dilation=dilation, bias=False),
        nn.BatchNorm2d(out_channels),
        nn.LeakyReLU(inplace=True, negative_slope=0.2),
    )


def BasicTransposeConv2d(in_channels, out_channels, kernel_size, stride, pad, dilation):
    output_pad = stride + 2 * pad - kernel_size * dilation + dilation - 1
    return nn.Sequential(
        nn.ConvTranspose2d(in_channels, out_channels, kernel_size, stride, pad, output_pad, dilation, bias=False),
        nn.BatchNorm2d(out_channels),
        nn.LeakyReLU(inplace=True, negative_slope=0.2),
    )


def BasicConv3d(in_channels, out_channels, kernel_size, stride, pad, dilation=1):
    return nn.Sequential(
        nn.Conv3d(in_channels, out_channels, kernel_size=kernel_size, stride=stride,
                  padding=pad, dilation=dilation, bias=False),
        nn.BatchNorm3d(out_channels),
        nn.LeakyReLU(inplace=True, negative_slope=0.2),
    )


def BasicTransposeConv3d(in_channels, out_channels, kernel_size, stride, pad, output_pad=0, dilation=1):
    # output_pad = stride + 2 * pad - kernel_size * dilation + dilation - 1
    return nn.Sequential(
        nn.ConvTranspose3d(in_channels, out_channels, kernel_size, stride,
                           pad, output_pad, dilation, bias=False),
        nn.BatchNorm3d(out_channels),
        nn.LeakyReLU(inplace=True, negative_slope=0.2),
    )


def conv3x3(in_planes: int, out_planes: int, stride: int = 1, groups: int = 1, dilation: int = 1) -> nn.Conv2d:
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)


def conv1x1(in_planes: int, out_planes: int, stride: int = 1) -> nn.Conv2d:
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class BasicBlock(nn.Module):
    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None
    ) -> None:
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out

class ConvBNReLU3d(nn.Sequential):
    def __init__(
        self,
        in_planes: int,
        out_planes: int,
        kernel_size: int = 3,
        stride: int = 1,
        groups: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None
    ) -> None:
        padding = (kernel_size - 1) // 2
        if norm_layer is None:
            norm_layer = nn.BatchNorm3d
        super(ConvBNReLU3d, self).__init__(
            nn.Conv3d(in_planes, out_planes, kernel_size, stride, padding, groups=groups, bias=False),
            norm_layer(out_planes),
            nn.ReLU6(inplace=True)
        )

class InvertedResidual3d(nn.Module):
    def __init__(
        self,
        inp: int,
        oup: int,
        stride: int,
        expand_ratio: int,
        norm_layer: Optional[Callable[..., nn.Module]] = None
    ) -> None:
        super(InvertedResidual3d, self).__init__()
        self.stride = stride
        assert stride in [1, 2]

        if norm_layer is None:
            norm_layer = nn.BatchNorm3d

        hidden_dim = int(round(inp * expand_ratio))
        self.use_res_connect = self.stride == 1 and inp == oup

        layers: List[nn.Module] = []
        if expand_ratio != 1:
            # pw
            layers.append(ConvBNReLU3d(inp, hidden_dim, kernel_size=1, norm_layer=norm_layer))
        layers.extend([
            # dw
            ConvBNReLU3d(hidden_dim, hidden_dim, stride=stride, groups=hidden_dim, norm_layer=norm_layer),
            # pw-linear
            nn.Conv3d(hidden_dim, oup, 1, 1, 0, bias=False),
            norm_layer(oup),
        ])
        self.conv = nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)

class AtrousBlock(nn.Module):

    def __init__(self, in_channels, out_channels, stride=1, bn=True, relu=True):
        super(AtrousBlock, self).__init__()

        dilations = [2,4,6]
        self.conv_1 = BasicConv(in_channels,out_channels//4,is_3d=True,kernel_size=3,stride=stride,padding=1,dilation=1)
        self.conv_2 = BasicConv(in_channels,out_channels//4,is_3d=True,kernel_size=3,stride=stride,padding=(1,dilations[0],dilations[0]),dilation=(1,dilations[0],dilations[0]))
        self.conv_3 = BasicConv(in_channels,out_channels//4,is_3d=True,kernel_size=3,stride=stride,padding=(1,dilations[1],dilations[1]),dilation=(1,dilations[1],dilations[1]))
        self.conv_4 = BasicConv(in_channels,out_channels//4,is_3d=True,kernel_size=3,stride=stride,padding=(1,dilations[2],dilations[2]),dilation=(1,dilations[2],dilations[2]))

    def forward(self, x):
        x = torch.cat((self.conv_1(x),self.conv_2(x),self.conv_3(x),self.conv_4(x)),1)
        return x
