from __future__ import print_function

import torch
import torch.nn as nn

from .util_conv import BasicConv
from .Submodule import SubModule

import pdb


class CostVolume(nn.Module):
    def __init__(self, maxdisp, cfg, glue=False, group=1):
        super(CostVolume, self).__init__()
        self.maxdisp = maxdisp+1
        self.mode = cfg['mode']
        self.glue = glue
        self.group = group
        self.unfold = nn.Unfold((1, maxdisp+1), 1, 0, 1)
        self.left_pad = nn.ZeroPad2d((maxdisp, 0, 0, 0))

    def forward(self, x, y, v=None):
        b, c, h, w = x.shape

        unfolded_y = self.unfold(self.left_pad(y)).reshape(
            b, self.group, c//self.group, self.maxdisp, h, w)
        x = x.reshape(b, self.group, c//self.group, 1, h, w)
        
        if self.mode == 'multiply':
            cost = (x*unfolded_y).sum(2)
        elif self.mode == 'subtract':
            cost = (x - unfolded_y).sum(2)
        else:
            raise NotImplementedError

        cost = torch.flip(cost, [2])

        if self.glue:
            cross = self.unfold(self.left_pad(v)).reshape(
                b, c, self.maxdisp, h, w)
            cross = torch.flip(cross, [2])
            return cost, cross
        else:
            return cost

class CostVolumeAA(nn.Module):
    def __init__(self, max_disp, feature_similarity='correlation'):
        """Construct cost volume based on different
        similarity measures

        Args:
            max_disp: max disparity candidate
            feature_similarity: type of similarity measure
        """
        super(CostVolumeAA, self).__init__()

        self.max_disp = max_disp
        self.feature_similarity = feature_similarity

    def forward(self, left_feature, right_feature):
        b, c, h, w = left_feature.size()

        if self.feature_similarity == 'difference':
            costs=[]
            for i in range(self.max_disp):
                if i > 0:
                    cost = (left_feature[:, :, :, i:] -
                                                right_feature[:, :, :, :-i]).mean(dim=1, keepdim=True)
                    cost = nn.functional.pad(cost, (i, 0, 0, 0), mode='constant')
                else:
                    cost = (left_feature - right_feature).mean(dim=1, keepdim=True)
                costs.append(cost)
            cost_volume = (torch.cat(costs, 1)).unsqueeze(1)

        elif self.feature_similarity == 'difference_gw':
            g = c // self.max_disp
            costs=[]
            for i in range(self.max_disp):
                s = i * g
                e = s + g
                if i > 0:
                    cost = (left_feature[:, s:e, :, i:] -
                                                right_feature[:, s:e, :, :-i]).mean(dim=1, keepdim=True)
                    cost = nn.functional.pad(cost, (i, 0, 0, 0), mode='constant')
                else:
                    cost = (left_feature[:, s:e:, :, :] - right_feature[:, s:e, :, :]).mean(dim=1,keepdim=True)
                costs.append(cost)
            cost_volume = (torch.cat(costs, 1)).unsqueeze(1)

        elif self.feature_similarity == 'concat':
            cost_volume = left_feature.new_zeros(b, 2 * c, self.max_disp, h, w)
            for i in range(self.max_disp):
                if i > 0:
                    cost_volume[:, :, i, :, i:] = torch.cat((left_feature[:, :, :, i:], right_feature[:, :, :, :-i]),
                                                            dim=1)
                else:
                    cost_volume[:, :, i, :, :] = torch.cat((left_feature, right_feature), dim=1)

        elif self.feature_similarity == 'correlation':
            cost_volume = left_feature.new_zeros(b, self.max_disp, h, w)

            for i in range(self.max_disp):
                if i > 0:
                    cost_volume[:, i, :, i:] = (left_feature[:, :, :, i:] *
                                                right_feature[:, :, :, :-i]).mean(dim=1)
                else:
                    cost_volume[:, i, :, :] = (left_feature * right_feature).mean(dim=1)
            cost_volume = cost_volume.unsqueeze(1)

        elif self.feature_similarity == 'correlation_gw':
            cost_volume = left_feature.new_zeros(b, self.max_disp, h, w)
            g = c // self.max_disp
            for i in range(self.max_disp):
                s = i * g
                e = s + g
                if i > 0:
                    cost_volume[:, i, :, i:] = (left_feature[:, s:e, :, i:] *
                                                right_feature[:, s:e, :, :-i]).mean(dim=1)
                else:
                    cost_volume[:, i, :, :] = (left_feature[:, s:e:, :, :] * right_feature[:, s:e, :, :]).mean(dim=1)
            cost_volume = cost_volume.unsqueeze(1)

        elif self.feature_similarity == 'correlation2':
            costs=[]
            for i in range(self.max_disp):
                if i > 0:
                    cost = (left_feature[:, :, :, i:] *
                                                right_feature[:, :, :, :-i]).mean(dim=1, keepdim=True)
                    cost = nn.functional.pad(cost, (i, 0, 0, 0), mode='constant')
                else:
                    cost = (left_feature * right_feature).mean(dim=1, keepdim=True)
                costs.append(cost)
            cost_volume = (torch.cat(costs, 1)).unsqueeze(1)

        elif self.feature_similarity == 'correlation2_gw':
            g = c // self.max_disp
            costs=[]
            for i in range(self.max_disp):
                s = i * g
                e = s + g
                if i > 0:
                    cost = (left_feature[:, s:e, :, i:] *
                                                right_feature[:, s:e, :, :-i]).mean(dim=1, keepdim=True)
                    cost = nn.functional.pad(cost, (i, 0, 0, 0), mode='constant')
                else:
                    cost = (left_feature[:, s:e:, :, :] * right_feature[:, s:e, :, :]).mean(dim=1,keepdim=True)
                costs.append(cost)
            cost_volume = (torch.cat(costs, 1)).unsqueeze(1)

        else:
            raise NotImplementedError

        cost_volume = cost_volume.contiguous()  # [B, C, D, H, W] or [B, D, H, W]

        return cost_volume


class AttentionCostVolume(nn.Module):
    def __init__(self, max_disparity, in_chan, hidden_chan, cfg, head=1, weighted=False):
        super(AttentionCostVolume, self).__init__()
        if cfg['type'] == 'corr_volume':
            self.costVolume = CostVolume(int(max_disparity//4), cfg, False, head)
        elif cfg['type'] == 'corr_volumeAA':
            self.costVolume = CostVolumeAA(int(max_disparity//4), cfg['mode'])
        self.conv = BasicConv(in_chan, hidden_chan, kernel_size=3, padding=1, stride=1)
        self.desc = nn.Conv2d(hidden_chan, hidden_chan, kernel_size=1, padding=0, stride=1)
        self.head = head
        self.weighted = weighted
        if weighted:
            self.weights = nn.Parameter(
                    torch.randn(hidden_chan).reshape(1, hidden_chan, 1, 1))

    def forward(self, imL, imR):
        b, _, h, w = imL.shape
        x = self.conv(imL)
        y = self.conv(imR)

        x_ = self.desc(x)
        y_ = self.desc(y)

        if self.weighted:
            weights = torch.sigmoid(self.weights)
            x_ = x_ * weights
            y_ = y_ * weights
        #cost = self.costVolume(x_/torch.norm(x_, 2, 1, True),
        #                       y_/torch.norm(y_, 2, 1, True))

        cost = self.costVolume(x_, y_)
        return cost


class disparityregression(nn.Module):
    def __init__(self):
        super(disparityregression, self).__init__()

    def forward(self, x, reg):
        out = torch.nn.functional.conv2d(x,reg)
        out = torch.squeeze(out,1)
        return out


class channelAtt(SubModule):
    def __init__(self, cv_chan, im_chan, D):
        super(channelAtt, self).__init__()

        self.im_att = nn.Sequential(
            BasicConv(im_chan, im_chan//2, kernel_size=1, stride=1, padding=0),
            nn.Conv2d(im_chan//2, cv_chan, 1))

        self.weight_init()

    def forward(self, cv, im):
        '''
        '''
        channel_att = self.im_att(im).unsqueeze(2)
        cv = torch.sigmoid(channel_att)*cv
        return cv
