import os
import sys
import numba
import pandas as pd
import numpy as np
import prettytable as pt
from collections import OrderedDict
import matplotlib.pyplot as plt
from tqdm import tqdm


def parser_result(res_file):
    """read and parser distance result to ordered dict
    """
    with open(res_file, 'r') as fin:
        res_dict = OrderedDict()
        for line in fin:
            # parser from line string
            data = line.strip().split('p[')[-1].split(']')[0]
            clip_name, _, timestamp_us, distances = data.split(':')
            res_dict.setdefault('clip_name', []).append(clip_name)
            res_dict.setdefault('timestamp_us', []).append(timestamp_us)

            for pair in distances.split(','):
                dist_name, dist = pair.strip().split(' ')
                res_dict.setdefault(dist_name, []).append(float(dist))
        return res_dict


def analysis_result(result):
    """analysis result using pandas lib
    """
    # pd.set_option('display.max_rows', None)
    df = pd.DataFrame(result, columns=result.keys())
    df[df.loc[:, df.columns[2:]] < 0.] = np.nan  # filter invalid value(-1)
    df = df[df.tof < 130.]
    df = df[df.tof > 0] # filter invalid tof case
    df = df[df.sgm > 0]
    
    # # print(df.describe())
    # print(df[df.sgm_err > 40])
    # exit(1)

    # calc abs rel for error
    sgm_abs_rel = pd.DataFrame(df.iloc[:, 4] / df.tof * 100,
                               columns=['sgm_abs_rel'])
    com_abs_rel = pd.DataFrame(df.iloc[:, 6] / df.tof * 100,
                               columns=[f'{df.columns[-2]}_abs_rel'])
    cat_df = pd.concat([df, sgm_abs_rel, com_abs_rel], axis=1)
    df = cat_df

    # print table
    tb = pt.PrettyTable()
    tb.field_names = ['distance'] + list(df.columns[2:])
    tb.align["distance"] = 'l'
    for min_d, max_d in [(0, 2), (2, 10), (10, 20), (20, 30), (30, 40),
                         (40, 50), (50, 60), (60, 100)]:
        sub_df = df[df.tof < max_d]
        sub_df = sub_df[sub_df.tof > min_d]
        field = f'{min_d}-{max_d}({len(sub_df.index)})'
        sub_mean = sub_df.describe().round(4).loc['mean']
        tb.add_row([field] + sub_mean.to_numpy().tolist())
    overall_mean = df.describe().round(4).loc['mean']
    tb.add_row([f'overall({len(df.index)})'] + overall_mean.to_numpy().tolist())

    print(tb.get_string(title="Stereo Distance Evaluation"))
    return df, tb

def images2html(imagelist):
    cols = 5
    video_urls = []
    for link_dir, image_path in imagelist:
        video_urls.append('<a href="{}"><img src="{}" height="200" width="300" /></a>'.format(link_dir, image_path))
        
    if len(video_urls) % cols != 0:
        add_num = cols - len(video_urls) % cols
        for i in range(add_num):
            video_urls.append('-') 
    video_array = np.array(video_urls)
    video_array = video_array.reshape(-1, cols)
    df = pd.DataFrame(video_array)
    return df.to_html(render_links=True, escape=False)
        
def plot_dataframe(df, table, save_dir):
    clip_names = sorted(set(df.clip_name.to_numpy().tolist()))
    fig_dir = os.path.join(save_dir, 'figures')
    if not os.path.exists(fig_dir):
        os.makedirs(fig_dir)
    figure_infos=[]
    for clip in tqdm(clip_names):
        clip_df = df[df.clip_name == clip]
        if len(clip_df.index) < 2: continue
        fig = clip_df.loc[:, df.columns[2:-2]].plot(title=clip).get_figure()
        plt.legend(loc='best')
        fig_name = os.path.join(fig_dir, clip + '.jpg')
        rel_path = os.path.join('figures', clip + '.jpg')
        fig.savefig(fig_name, format='jpg', transparent=True, dpi=150, pad_inches = 0)
        figure_infos.append((rel_path, rel_path))
        plt.close(fig)
        
    with open(save_dir + '/stereo_distance_eval.html', 'w') as fou:
        fou.write('<meta http-equiv="Content-Type" content="text/html; charset=utf-8">')
        fou.write('<h1 align="center">Stereo Distance Evaluation</h1>\n')         
        fou.write('<h2>评测结果</h2>\n')
        fou.write(table.get_html_string(title="Stereo Distance Evaluation", attributes={"border": 1, "style": "border-width: 1px; border-collapse: collapse;"}))
        fou.write('<h2>曲线图</h2>\n')
        fou.write(images2html(figure_infos))
        
        

def save2csv(df, save_csv_name):
    df.to_csv(save_csv_name)
    
    
def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
            It must be "yes" (the default), "no" or None (meaning
            an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")

if __name__ == '__main__':
    res_file = sys.argv[1]
    res = parser_result(res_file)
    df, table = analysis_result(res)
    save2csv(df, res_file + '.csv')
    
    if query_yes_no('Do you want to plot figures?'):
        plot_dataframe(df, table, os.path.splitext(os.path.basename(res_file))[0])
