import os
import sys
import shutil
import glob
import numpy as np
import cv2
import time
import skimage
import skimage.io

import torch
from torch.nn import functional as F
from torch.utils.data import DataLoader
import pytorch_lightning as pl
from pytorch_lightning.core.lightning import LightningModule
from pytorch_lightning.loggers import TestTubeLogger
from pytorch_lightning.utilities import rank_zero_only

from ruamel.yaml import YAML
from loguru import logger as logur

# from models import *
from utils.load import load_class
from dataloaders import listflowfile as lt
from dataloaders import SceneFlowLoader as SFL
from dataloaders import KITTIloader2012 as ls12
from dataloaders import KITTIloader2015 as ls15
from dataloaders import KITTILoader as KL
from dataloaders import KITTI_submission_loader as KSL

import pdb
import warnings

warnings.filterwarnings('ignore')


configs = [
    'cfg_coex.yaml',
    'cfg_psm.yaml'
    ]
config_num = 0


class Stereo(LightningModule):

    def __init__(self, cfg, dataname=None):
        super().__init__()
        self.cfg = cfg
        self.dataname = dataname

        self.stereo = load_class(
            cfg['model']['stereo']['name'],
            ['models.stereo'])(cfg['model']['stereo'])


    #def forward(self, imgLR, training=False):
        #imgL, imgR = torch.chunk(imgLR, 2, 1)
    def forward(self, imgL, imgR, training=False):
        if training:
            disp_pred = self.stereo(imgL, imgR, training=training)
        else:
            h, w = imgL.shape[-2:]
            h_pad = 32-h % 32
            w_pad = 32-w % 32

            #imgL = F.pad(imgL, (0, w_pad, 0, h_pad))
            #imgR = F.pad(imgR, (0, w_pad, 0, h_pad))

            disp_pred = self.stereo(imgL, imgR, training=training)[0][:, :h, :w]
            
        return disp_pred

    def training_step(self, batch, batch_idx):
        imgL, imgR, disp_true = batch['imgL'], batch['imgR'], batch['disp_true']
        x1, y1 = batch['x1'], batch['y1']
        disp_pred = self.stereo(imgL, imgR, u0=x1, v0=y1, training=True)

        losses = []
        train_weights = self.cfg['training']['training_scales_weighting']
        for disp_pred_ in disp_pred[:-1]:

            disp_pred_ = F.interpolate(
                disp_pred_.unsqueeze(1),
                size=(disp_true.shape[1], disp_true.shape[2]),
                mode='bilinear').squeeze(1)
            # ---------
            if self.dataname == 'sceneflow':
                mask = torch.logical_and(
                    disp_true <= self.cfg['model']['stereo']['max_disparity'],
                    disp_true > 0)
            else:
                mask = disp_true > 0.001
            mask.detach_()
            if (disp_pred_[mask]).numel() == 0:
                return None
            # ----
            losses.append(F.smooth_l1_loss(
                disp_pred_[mask], disp_true[mask], reduction='mean'))
        loss = sum([losses[i] * train_weights[i] for i in range(len(disp_pred[:-1]))]) /\
            sum([1 * train_weights[i] for i in range(len(disp_pred[:-1]))])

        self.log('{}_train_loss'.format(self.dataname), loss,
                 on_step=True, on_epoch=True, prog_bar=True, logger=True)
        return loss

    def validation_step(self, batch, batch_idx):
        imgL, imgR, disp_true = batch['imgL'], batch['imgR'], batch['disp_true']

        h, w = imgL.shape[-2:]
        h_pad = (32-(h % 32)) % 32
        w_pad = (32-(w % 32)) % 32

        imgL = F.pad(imgL, (0, w_pad, 0, h_pad))
        imgR = F.pad(imgR, (0, w_pad, 0, h_pad))

        disp_pred = self.stereo(imgL, imgR)[0][:, :h, :w]

        # ---------
        mask = torch.logical_and(disp_true <= self.cfg['model']['stereo']['max_disparity'],
                                 disp_true > 0)
        mask.detach_()
        # ----
        if mask.sum() == 0:
            epe = 0
        else:
            epe = torch.mean(torch.abs(disp_pred[mask]-disp_true[mask])).detach()
            if torch.isnan(epe).sum() > 0:
                pdb.set_trace()

        # ---------
        mask = disp_true > 0.001
        mask.detach_()
        # ----
        if mask.sum() == 0:
            error_3px, error_2px, error_1px = 0, 0, 0
        else:
            delta = torch.abs(disp_pred[mask] - disp_true[mask])
            error_mat_3px = (torch.logical_and((delta >= 3.0), (delta >= 0.05 * (disp_true[mask]))) == True)
            error_3px = torch.sum(error_mat_3px).item() / torch.numel(disp_pred[mask]) * 100
            error_mat_2px = (torch.logical_and((delta >= 2.0), (delta >= 0.05 * (disp_true[mask]))) == True)
            error_2px = torch.sum(error_mat_2px).item() / torch.numel(disp_pred[mask]) * 100
            error_mat_1px = (torch.logical_and((delta >= 1.0), (delta >= 0.05 * (disp_true[mask]))) == True)
            error_1px = torch.sum(error_mat_1px).item() / torch.numel(disp_pred[mask]) * 100

        self.log_dict(
            {
                '{}_val_epe'.format(self.dataname): epe,
                '{}_val_3pxError'.format(self.dataname): error_3px,
                '{}_val_2pxError'.format(self.dataname): error_2px,
                '{}_val_1pxError'.format(self.dataname): error_1px
            }
        )

        disp_img = disp_pred[0].data.cpu().numpy().astype(np.uint8)
        disp_img = cv2.applyColorMap(disp_img*2, cv2.COLORMAP_JET)
        self.save_disp_imgs(disp_img, '{:05d}'.format(batch_idx))
        return epe

    def test_step(self, batch, batch_idx):
        if self.dataname == 'sceneflow':
            imgL, imgR, disp_true = batch['imgL'], batch['imgR'], batch['disp_true']
            dataname, filename = None, '{:05d}'.format(batch_idx)
        else:
            imgL, imgR = batch['imgL'], batch['imgR']
            dataname, filename = batch['dataname'][0], batch['filename'][0]

        h, w = imgL.shape[-2:]
        h_pad = (32-(h % 32)) % 32
        w_pad = (32-(w % 32)) % 32

        imgL = F.pad(imgL, (0, w_pad, 0, h_pad))
        imgR = F.pad(imgR, (0, w_pad, 0, h_pad))

        start = torch.cuda.Event(enable_timing=True)
        end = torch.cuda.Event(enable_timing=True)
        start.record()

        disp_pred = self.stereo(imgL, imgR)[0][:, :h, :w]
        end.record()
        torch.cuda.synchronize()
        runtime = start.elapsed_time(end)
        fps = 1000/runtime
        # print('total runtime : ', runtime)

        if self.dataname == 'sceneflow':
            if self.cfg['testing']['compute_metrics']:
                # mask for every image
                for left_img, disp_pr, disp_tr, left_path in zip(imgL, disp_pred, disp_true, batch['left_path']):
                    filename = os.path.splitext(os.path.basename(left_path))[0]
                    # ---------
                    mask = torch.logical_and(disp_tr <= self.cfg['model']['stereo']['max_disparity'],
                                         disp_tr > 0)
                    mask.detach_()
                    # ----
                    if mask.sum() == 0:
                        epe = 0
                    else:
                        epe = torch.mean(torch.abs(disp_pr[mask]-disp_tr[mask])).detach()
                        if torch.isnan(epe).sum() > 0:
                            pdb.set_trace()

                    # ---------
                    #mask = disp_true > 0.001
                    #mask.detach_()
                    # ----
                    if mask.sum() == 0:
                        error_3px, error_2px, error_1px = 0, 0, 0
                        logur.warning(f'Warning! Ignore {left_path} because it is invalid.')
                        continue 
                    else:
                        delta = torch.abs(disp_pr - disp_tr)
                        error_mat_3px = torch.logical_and(delta >= 3.0, mask)
                        error_3px = torch.sum(error_mat_3px).item() / torch.numel(disp_pr[mask]) * 100
                        error_mat_2px = torch.logical_and(delta >= 2.0, mask)
                        error_2px = torch.sum(error_mat_2px).item() / torch.numel(disp_pr[mask]) * 100
                        error_mat_1px = torch.logical_and(delta >= 1.0, mask)
                        error_1px = torch.sum(error_mat_1px).item() / torch.numel(disp_pr[mask]) * 100
                        error_mat_05px = torch.logical_and(delta >= 0.5, mask)
                        error_05px = torch.sum(error_mat_05px).item() / torch.numel(disp_pr[mask]) * 100
                        error_mat_01px = torch.logical_and(delta >= 0.1, mask)
                        error_01px = torch.sum(error_mat_01px).item() / torch.numel(disp_pr[mask]) * 100
                        
                        res_dict = {'test_epe': epe.item(),
                               'test_3pxError': error_3px,
                               'test_2pxError': error_2px,
                               'test_1pxError': error_1px,
                               'test_05pxError': error_05px,
                               'test_01pxError': error_01px,
                               'test_fps': fps}
                        self.log_dict(res_dict)
                        logur.info({filename:res_dict})
                        
                        if self.cfg['testing']['save_disp_imgs']:
                            left_img = left_img.permute(1, 2, 0).cpu().numpy()
                            left_img = (left_img - left_img.min()) / (left_img.max() - left_img.min()) * 255
                            left_img = left_img.astype(np.uint8)

                            disp_gt = disp_tr.data.cpu().numpy()
                            disp_img = disp_pr.data.cpu().numpy()

                            disp_min = min(disp_gt.min(), disp_img.min())
                            disp_max = max(disp_gt.max(), disp_img.max())

                            disp_gt_norm = (disp_gt - disp_min) / (disp_max - disp_min) * 255
                            disp_gt_color = cv2.applyColorMap(disp_gt_norm.astype(np.uint8), cv2.COLORMAP_JET)

                            disp_img_norm = (disp_img - disp_min) / (disp_max - disp_min) * 255
                            disp_img_color = cv2.applyColorMap(disp_img_norm.astype(np.uint8), cv2.COLORMAP_JET)

                            cat_disp_img = np.concatenate((left_img, disp_img_color, disp_gt_color), axis=1)
                            for i, (tag, error) in enumerate([('error_1px', error_1px), ('error_0.5px', error_05px), ('error_0.1px', error_01px)]):
                                cv2.putText(cat_disp_img, f'{tag}: {error:.4f}', (20, 30*(i+1)), cv2.FONT_HERSHEY_COMPLEX, 1, (255, 255, 255))
                            # self.save_disp_imgs(save_img, filename, dataname)

                            cat_error_img = (torch.cat((error_mat_1px, error_mat_05px, error_mat_01px), axis=1)*255).squeeze().cpu().numpy().astype(np.uint8)
                            cat_error_img = np.expand_dims(cat_error_img, 2).repeat(3,axis=2)
                            cat_img = np.concatenate((cat_disp_img, cat_error_img), axis=0)
                            self.save_disp_imgs(cat_img, filename, dataname)
                        
                return None
            else:
                self.log_dict({'{}_test_fps'.format(self.dataname): fps})
                return {'{}_fps'.format(self.dataname): fps}

        else:
            if self.cfg['testing']['save_disp_imgs']:
                disp_img = (256*disp_pred[0].data.cpu().numpy()).astype('uint16')
                self.save_disp_imgs(disp_img, filename, dataname, True)

            self.log_dict({'{}_test_fps'.format(self.dataname): fps})

            return {'{}_fps'.format(self.dataname): fps}

    def configure_optimizers(self):
        dataname = 'kitti' if 'kitti' in self.dataname else 'sceneflow'
        optimizer = torch.optim.Adam(
            self.parameters(), lr=self.cfg['training']['lr'], betas=(0.9, 0.999))
        lr_scheduler = torch.optim.lr_scheduler.MultiStepLR(
            optimizer,
            milestones=self.cfg['training']['{}_milestones'.format(dataname)],
            gamma=self.cfg['training']['{}_gamma'.format(dataname)])
        return [optimizer], [lr_scheduler]

    @rank_zero_only
    def make_dirs(self, dirname):
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    def save_disp_imgs(self, disp_img, filename,
                       dataname=None, use_skimage=False):
        # Save imgs
        if dataname is None:
            dataname = self.dataname
        savedir = '{}/{}/version_{}/media/{}'.format(
            self.logger.save_dir, self.logger.name,
            self.logger.version, dataname)
        self.make_dirs(savedir)

        if use_skimage:
            skimage.io.imsave('{}/{}/version_{}/media/{}/{}.jpg'.format(
                self.logger.save_dir, self.logger.name,
                self.logger.version, dataname, filename), disp_img)
        else:
            cv2.imwrite('{}/{}/version_{}/media/{}/{}.jpg'.format(
                self.logger.save_dir, self.logger.name,
                self.logger.version, dataname, filename), disp_img)


def load_configs(path):
    cfg = YAML().load(open(path, 'r'))
    backbone_cfg = YAML().load(
        open(cfg['model']['stereo']['backbone']['cfg_path'], 'r'))
    cfg['model']['stereo']['backbone'].update(backbone_cfg)
    return cfg


def copy_dir(save_dir, name, save_version):
    savedir = '{}/{}/version_{}/project/'.format(save_dir, name, save_version)
    datadirs = ['configs', 'dataloaders', 'models', 'utils']
    if os.path.exists(savedir):
        shutil.rmtree(savedir)
    os.makedirs(savedir)
    for file in glob.glob('./*.py'):
        shutil.copyfile('./{}'.format(file),
                        '{}{}'.format(savedir, file))
    for datadir in datadirs:
        shutil.copytree('./{}'.format(datadir),
                        '{}{}'.format(savedir, datadir))
    return save_version

if __name__ == '__main__':
    pl.seed_everything(42)
    cfg = load_configs(sys.argv[1])
    eval_ckpt = None
    if len(sys.argv) > 2 and 'ckpt' in sys.argv[2]:
        eval_ckpt = sys.argv[2]
        print(f'Just eval {eval_ckpt}')
    # os.environ['CUDA_VISIBLE_DEVICES'] = cfg['device']
    logging_pth = cfg['training']['paths']['logging']
    ###
    th, tw = cfg['training']['th'], cfg['training']['tw']
    ''' SceneFlow Training Part '''
    if cfg['training']['train_on']['sceneflow']:
        sceneflowpath = cfg['training']['paths']['sceneflow']
        all_left_img, all_right_img, all_left_disp, all_focal, test_left_img, test_right_img, test_left_disp, test_focal = lt.dataloader(sceneflowpath)
        sceneflow_train = SFL.ImageLoader(
            all_left_img, all_right_img, all_focal, all_left_disp,
            True, th=th, tw=tw)
        sceneflow_train = DataLoader(
            sceneflow_train, batch_size=cfg['training']['batch_size'],
            num_workers=16, shuffle=True, drop_last=False)
        sceneflow_test = SFL.ImageLoader(
            test_left_img, test_right_img, test_focal, test_left_disp,
            False)
        sceneflow_test = DataLoader(
            sceneflow_test, batch_size=1, num_workers=16,
            shuffle=False, drop_last=False)

        # Model
        log_name = 'sceneflow'
        if cfg['training']['load_version'] is not None:
            load_version = cfg['training']['load_version']
            ckpt = '{}/{}/version_{}/checkpoints/sceneflow-epoch={}.ckpt'.format(
                logging_pth, cfg['model']['name'], load_version,
                cfg['training']['sceneflow_max_epochs']-1)
            stereo = Stereo.load_from_checkpoint(ckpt, cfg=cfg, dataname=log_name)
            
        else:
            stereo = Stereo(cfg, 'sceneflow')

        version = copy_dir(
            logging_pth, cfg['model']['name'], cfg['training']['save_version'])
        # resume_from_checkpoint = None

        logger = TestTubeLogger(
            logging_pth,
            cfg['model']['name'],
            version=version)
        gpu_stats = pl.callbacks.GPUStatsMonitor()
        lr_monitor = pl.callbacks.LearningRateMonitor()
        checkpoint_callback = pl.callbacks.ModelCheckpoint(
            filename=log_name+'-{epoch}',
            save_last=True,
            save_top_k=-1,
            every_n_epochs=1,
            monitor=log_name+'_train_loss_epoch')
        trainer = pl.Trainer(
            accelerator='dp',
            logger=logger,
            callbacks=[gpu_stats, lr_monitor, checkpoint_callback],
            precision=cfg['precision'],
            gpus=cfg['device'],
            max_epochs=cfg['training']['sceneflow_max_epochs'],
            # resume_from_checkpoint=resume_from_checkpoint,
            benchmark=True,
            accumulate_grad_batches=1,
            gradient_clip_val=0.1,
            stochastic_weight_avg=True,
            # track_grad_norm=2,
            weights_summary='full',
            )

        trainer.fit(stereo, sceneflow_train)
        trainer.test(stereo, sceneflow_test)

    ''' Custom Training Part '''
    if cfg['training']['train_on']['custom']:
        train_path = cfg['training']['paths']['custom_train']
        train_list = cfg['training']['paths']['train_list']
        test_path = cfg['training']['paths']['custom_test']
        test_list = cfg['training']['paths']['test_list']
        from dataloaders.stereo import listcustom as ls
        from dataloaders.stereo import CustomLoader as DA
        all_left_img, all_right_img, all_left_disp, test_left_img, test_right_img, test_left_disp = ls.dataloader(train_path, train_list, test_path, test_list)
        TrainImgLoader = torch.utils.data.DataLoader(
            DA.myImageFloder(all_left_img,all_right_img,all_left_disp, True, th=th, tw=tw), 
            batch_size=cfg['training']['batch_size'], shuffle=True, num_workers=16, drop_last=False, pin_memory=True)

        TestImgLoader = torch.utils.data.DataLoader(
            DA.myImageFloder(test_left_img,test_right_img,test_left_disp, False, th=th, tw=tw), 
            batch_size=4, shuffle= False, num_workers=4, drop_last=False)


        # Model
        log_name = 'sceneflow'
        if cfg['training']['load_version'] is not None:
            load_version = cfg['training']['load_version']
            ckpt = load_version
            if eval_ckpt:
                ckpt = eval_ckpt
            print(f'Loading ckpt from :{ckpt}')
            stereo = Stereo.load_from_checkpoint(ckpt, cfg=cfg, dataname=log_name)
            
        else:
            stereo = Stereo(cfg, log_name)

        version = copy_dir(
            logging_pth, cfg['model']['name'], cfg['training']['save_version'])
        # resume_from_checkpoint = None

        logger = TestTubeLogger(
            logging_pth,
            cfg['model']['name'],
            version=version)
        logfile = '{}/{}/version_{}/{}.log'.format(
            logger.save_dir, logger.name,
            logger.version, time.strftime("%Y-%m-%d_%H-%M-%S", time.localtime()))
        logur.add(logfile)
        gpu_stats = pl.callbacks.GPUStatsMonitor()
        lr_monitor = pl.callbacks.LearningRateMonitor()
        checkpoint_callback = pl.callbacks.ModelCheckpoint(
            filename=log_name+'-{epoch}',
            save_last=True,
            save_top_k=5,
            mode='min',
            every_n_epochs=1,
            monitor=log_name+'_train_loss_epoch')
        trainer = pl.Trainer(
            accelerator='ddp',
            logger=logger,
            callbacks=[gpu_stats, lr_monitor, checkpoint_callback],
            precision=cfg['precision'],
            gpus=cfg['device'] if torch.cuda.device_count() >=len(cfg['device']) else [0],
            max_epochs=cfg['training']['sceneflow_max_epochs'],
            # resume_from_checkpoint=resume_from_checkpoint,
            benchmark=True,
            accumulate_grad_batches=1,
            gradient_clip_val=0.1,
            stochastic_weight_avg=True,
            # track_grad_norm=2,
            weights_summary='full',
            )

        if eval_ckpt is None:
            trainer.fit(stereo, TrainImgLoader)
        res = trainer.test(stereo, TestImgLoader)
        logur.info({'Test_Result': res[0]})



    ''' KITTI Training Part '''
    all_left_img, all_right_img, all_left_disp, all_calib, val_left_img, val_right_img, val_left_disp, val_calib = [], [], [], [], [], [], [], []
    if cfg['training']['train_on']['kitti12']:
        all_left_img12, all_right_img12, all_left_disp12, all_calib12, val_left_img12, val_right_img12, val_left_disp12, val_calib12 = ls12.dataloader(
            cfg['training']['paths']['kitti12'], True)
        all_left_img += all_left_img12
        all_right_img += all_right_img12
        all_left_disp += all_left_disp12
        all_calib += all_calib12
        val_left_img += val_left_img12
        val_right_img += val_right_img12
        val_left_disp += val_left_disp12
        val_calib += val_calib12
        log_name = 'kitti12'

    if cfg['training']['train_on']['kitti15']:
        all_left_img15, all_right_img15, all_left_disp15, all_calib15, val_left_img15, val_right_img15, val_left_disp15, val_calib15 = ls15.dataloader(
            cfg['training']['paths']['kitti15'], True)
        all_left_img += all_left_img15
        all_right_img += all_right_img15
        all_left_disp += all_left_disp15
        all_calib += all_calib15
        val_left_img += val_left_img15
        val_right_img += val_right_img15
        val_left_disp += val_left_disp15
        val_calib += val_calib15
        log_name = 'kitti15'

    if cfg['training']['train_on']['kitti12'] and cfg['training']['train_on']['kitti15']:
        log_name = 'kitti'

    if cfg['training']['train_on']['kitti12'] or cfg['training']['train_on']['kitti15']:
        kitti_train = KL.ImageLoader(
            all_left_img, all_right_img, all_left_disp, all_calib, th, tw,
            training=True)
        kitti_train = DataLoader(
            kitti_train, batch_size=cfg['training']['batch_size'],
            num_workers=16, shuffle=True, drop_last=False)
        kitti_val = KL.ImageLoader(
            val_left_img, val_right_img, val_left_disp, val_calib,
            training=False)
        kitti_val = DataLoader(
            kitti_val, batch_size=1,
            num_workers=16, shuffle=False, drop_last=False)

        # Model
        if cfg['training']['load_version'] is not None:
            load_version = cfg['training']['load_version']
        else:
            assert(cfg['training']['train_on']['sceneflow'])
            load_version = cfg['training']['save_version']

        ckpt = '{}/{}/version_{}/checkpoints/sceneflow-epoch={}.ckpt'.format(
            logging_pth, cfg['model']['name'], load_version,
            cfg['training']['sceneflow_max_epochs']-1)
        # ckpt = '{}/{}/version_{}/checkpoints/last.ckpt'.format(
        #     logging_pth, cfg['model']['name'], load_version)
        stereo = Stereo.load_from_checkpoint(ckpt, cfg=cfg, dataname=log_name)

        logger = TestTubeLogger(
            logging_pth,
            cfg['model']['name'],
            version=cfg['training']['save_version'])
        gpu_stats = pl.callbacks.GPUStatsMonitor()
        lr_monitor = pl.callbacks.LearningRateMonitor()
        checkpoint_callback = pl.callbacks.ModelCheckpoint(
            filename=log_name+'-{epoch}',
            save_last=True,
            save_top_k=2,
            monitor=log_name+'_val_3pxError')
        trainer = pl.Trainer(
            accelerator='dp',
            logger=logger,
            callbacks=[gpu_stats, lr_monitor, checkpoint_callback],
            precision=cfg['precision'],
            gpus=cfg['device'],
            max_epochs=cfg['training']['kitti_max_epochs'],
            # resume_from_checkpoint=resume_from_checkpoint,
            benchmark=True,
            accumulate_grad_batches=1,
            gradient_clip_val=0.1,
            stochastic_weight_avg=True,
            # track_grad_norm=2,
            weights_summary='full',
            )

        trainer.fit(stereo, kitti_train, kitti_val)

    if cfg['training']['train_on']['kitti12']:
        datapath = cfg['training']['paths']['kitti12'].replace('training', 'testing')
        test_left_img, test_right_img, test_calib = KSL.listfiles(datapath, 'kitti12')
        kitti_test = KSL.ImageLoader(test_left_img, test_right_img, test_calib)
        kitti_test = DataLoader(
            kitti_test, batch_size=1,
            num_workers=16, shuffle=False, drop_last=False)
        trainer.test(stereo, kitti_test)

    if cfg['training']['train_on']['kitti15']:
        datapath = cfg['training']['paths']['kitti15'].replace('training','testing')
        test_left_img, test_right_img, test_calib = KSL.listfiles(datapath, 'kitti15')
        kitti_test = KSL.ImageLoader(test_left_img, test_right_img, test_calib)
        kitti_test = DataLoader(
            kitti_test, batch_size=1,
            num_workers=16, shuffle=False, drop_last=False)
        trainer.test(stereo, kitti_test)
